import functools
from functools import wraps
from typing import Optional, Callable

from flask import request, jsonify
from flask_cors import cross_origin

from bot_api.abc.auth_service_abc import AuthServiceABC
from bot_api.exception.service_error_code_enum import ServiceErrorCode
from bot_api.exception.service_exception import ServiceException
from bot_api.model.error_dto import ErrorDTO
from bot_data.abc.auth_user_repository_abc import AuthUserRepositoryABC
from bot_data.model.auth_role_enum import AuthRoleEnum


class Route:
    registered_routes = {}

    _auth_users: Optional[AuthUserRepositoryABC] = None
    _auth: Optional[AuthServiceABC] = None

    @classmethod
    def init_authorize(cls, auth_users: AuthUserRepositoryABC, auth: AuthServiceABC):
        cls._auth_users = auth_users
        cls._auth = auth

    @classmethod
    def authorize(cls, f: Callable = None, role: AuthRoleEnum = None):
        if f is None:
            return functools.partial(cls.authorize, role=role)

        @wraps(f)
        async def decorator(*args, **kwargs):
            token = None
            if 'Authorization' in request.headers:
                bearer = request.headers.get('Authorization')
                token = bearer.split()[1]

            if token is None:
                ex = ServiceException(ServiceErrorCode.Unauthorized, f'Token not set')
                error = ErrorDTO(ex.error_code, ex.message)
                return jsonify(error.to_dict()), 401

            if cls._auth_users is None or cls._auth is None:
                ex = ServiceException(ServiceErrorCode.Unauthorized, f'Authorize is not initialized')
                error = ErrorDTO(ex.error_code, ex.message)
                return jsonify(error.to_dict()), 401

            if not cls._auth.verify_login(token):
                ex = ServiceException(ServiceErrorCode.Unauthorized, f'Token expired')
                error = ErrorDTO(ex.error_code, ex.message)
                return jsonify(error.to_dict()), 401

            token = cls._auth.decode_token(token)
            if token is None or 'email' not in token:
                ex = ServiceException(ServiceErrorCode.Unauthorized, f'Token invalid')
                error = ErrorDTO(ex.error_code, ex.message)
                return jsonify(error.to_dict()), 401

            user = cls._auth_users.get_auth_user_by_email(token['email'])
            if user is None:
                ex = ServiceException(ServiceErrorCode.Unauthorized, f'Token invalid')
                error = ErrorDTO(ex.error_code, ex.message)
                return jsonify(error.to_dict()), 401

            if role is not None and user.auth_role.value < role.value:
                ex = ServiceException(ServiceErrorCode.Unauthorized, f'Role {role} required')
                error = ErrorDTO(ex.error_code, ex.message)
                return jsonify(error.to_dict()), 403

            return await f(*args, **kwargs)

        return decorator

    @classmethod
    def route(cls, path=None, **kwargs):
        # simple decorator for class based views
        def inner(fn):
            cross_origin(fn)
            cls.registered_routes[path] = (fn, kwargs)
            return fn

        return inner

    @classmethod
    def get(cls, path=None, **kwargs):
        return cls.route(path, methods=['GET'], **kwargs)

    @classmethod
    def post(cls, path=None, **kwargs):
        return cls.route(path, methods=['POST'], **kwargs)

    @classmethod
    def head(cls, path=None, **kwargs):
        return cls.route(path, methods=['HEAD'], **kwargs)

    @classmethod
    def put(cls, path=None, **kwargs):
        return cls.route(path, methods=['PUT'], **kwargs)

    @classmethod
    def delete(cls, path=None, **kwargs):
        return cls.route(path, methods=['DELETE'], **kwargs)
